/*  
    Copyright 2007, Francois Schiettekatte

    This file is part of Corteo.

    Corteo is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Corteo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Corteo.  If not, see <http://www.gnu.org/licenses/>.

*/

#include <stdio.h>
#include <math.h>

#include "corteoindex.h"
#include "corteomatrix.h"
#include "constants.h"

// scattering matrix: sin^2(theta_CM/2)
float matrix[DIME*DIMS];

#define THETAERR -1000.0

// parameters used to compute matrix, stored in matrix file
#define HEADERSIZE 7
float headerRef[HEADERSIZE] = {MINE, DIME, MAXE, MINS, DIMS, MAXS, NSUM}; 


// insert your prefered screening function here
double PHI(double x) {
  // Universal
  return 0.1818*exp(-3.*x)+0.5099*exp(-0.9423*x)+0.2802*exp(-0.4028*x)+0.02817*exp(-0.2016*x);
}

// Following functions used to compute scattering angle following Gauss-Mehler quadrature
double H(double u, double x0, double epsilon, double s) {
  double x0u = x0/u;
  double eps1 = 1./epsilon;
  double ss = s*s;
  return sqrt((1-u*u)/(1-PHI(x0u)/x0u*eps1-ss/x0u/x0u));
}

// function to minimize
double funcX0(double x0, double epsilon, double s) {
    double sx0 = s/x0;
    return 1.-PHI(x0)/(x0*epsilon)-sx0*sx0;
}

// use bisection method to find x0
// (return THETAERR on error)
double findX0(double epsilon, double s) {
   double x1 = MINS;
   double x2 = MAXS;
   double funcX0val1 = funcX0(x1, epsilon, s);  // should be always negative
   double funcX0val2 = funcX0(x2, epsilon, s);  // should be always positive (starting ~1.0)
   double funcX0valm = funcX0((x1+x2)/2., epsilon, s);

   if(funcX0val1>=0. || funcX0val2<=0.)
     return THETAERR;    // error, values should be on each side of 0

   do {
     if(funcX0valm<0.) {
       funcX0val1 = funcX0valm;
       x1 = (x1+x2)/2.;
     } else {
       funcX0val2 = funcX0valm;
       x2 = (x1+x2)/2.;
     }
     funcX0valm = funcX0((x1+x2)/2., epsilon, s);
   } while (fabs(funcX0valm)>1e-14);
    // 1.-XXX wont be more precise than 2^(-52)=2e-16
    // but using 1e-10 saves time and is enough precise for float conversion later on

   return (x1+x2)/2.;
}

// scattering angle calculated using Gauss-Mehler quadrature
// (return THETAERR on error)
double THETA(double epsilon, double s, unsigned int nsum) {
  double temp, sum = 0.;
  double x0 = findX0(epsilon, s);
  unsigned int j;

  if(x0==THETAERR) return THETAERR;

  for(j=0; j<nsum/2; j++) {
    temp = H(cos((2.*j-1.)*PI/(2.*nsum)),x0, epsilon, s);
    sum += temp;
  }

  return PI - 2.*PI*s/nsum/x0*sum;

}

// compute all the elements of matrix and write matrix to file 'corteo.mat'
// user sets showProgress!=0 to display the progress of this (long) calculation to the console
// return 1 if successful, 0 if not able to write file
int calcMatrix(int showProgress) {
  unsigned long i, j, nThetaErr = 0;
  double theta, sinThetaBy2;

  FILE *ofp;
  char *mode = "wb";

  if(showProgress) fprintf(stdout, "Computing scattering matrix");
  // compute matrix for each reduced energy, reduced impact parameter pair
  for(j=0; j<DIME; j++) {
    if(showProgress && j%(DIME/10)==0) {
      fprintf(stdout, ".");
      fflush(stdout);
    }

    for(i=0; i<DIMS; i++) {
       // calculations (summation) made using double to decrease numerical noise
       theta = THETA(Eval(j), Sval(i), NSUM);
       if(theta == THETAERR) nThetaErr++;
       sinThetaBy2 = sin(theta/2.);

       // store in matrix sin^2(theta_CM/2) as float 
       setMatrix(i*DIME+j,(float)(sinThetaBy2*sinThetaBy2));
    }
  }

  // open file corteo.mat and write matrix, including a header indicating the parameters used to compute matrix
  ofp = fopen("data/corteo.mat", mode);
  if (ofp == NULL) return 0;
  fwrite((void*)headerRef, HEADERSIZE, sizeof(float), ofp);
  fwrite((void*)matrix,    DIME*DIMS,  sizeof(float), ofp);
  fclose(ofp);

  if(showProgress) { 
    fprintf(stdout, "done,\n");
    fflush(stdout);
  }
  if(nThetaErr) fprintf(stderr, "%u error(s) evaluating theta\n", nThetaErr);
  
  return 1;
}


// read scattering matrix from file corteo.mat
// return 1 if file read correctly, 0 otherwise
int loadMatrix(void) {
  unsigned int i;
  FILE *ifp;
  char *mode = "rb";

  float header[HEADERSIZE];
  unsigned int tail;

  ifp = fopen("data/corteo.mat", mode);
  if (ifp == NULL) {
    fprintf(stderr,"Can't open data/corteo.mat ");
    return 0;
  }

  // file exists, get header
  fread((void*)header, HEADERSIZE, sizeof(float), ifp);

  for(i=0; i<HEADERSIZE; i++)
    if(header[i] != headerRef[i]) {
	    // "corteo.mat" header does not correspond to currently defined parameters
	    fclose(ifp);
      fprintf(stderr,"Header element %i does not correspond to current parameters in corteo.mat",i);
      return 0;
    }

  // read file into matrix
  fread((void*)matrix, DIME*DIMS, sizeof(float), ifp);

  // read control data
  fread((void*)(&tail), 1, sizeof(int), ifp);
  fclose(ifp);

  return 1;
}


// return element i of matrix
float Matrix(unsigned long i) { 
	return matrix[i]; 
}

// set element i of matrix
void setMatrix(unsigned long i, float val) {
	matrix[i] = val; 
}
