/*  
    Copyright 2007, Francois Schiettekatte

    This file is part of Corteo.

    Corteo is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Corteo is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Corteo.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "corteoutil.h"
#include "randomx.h"

#include <math.h>
#include <float.h>
#include <stdlib.h> // for correct definition of atof()

// randomly swap 3 times each elements of list so they come out in random order
void randomizelist(float *list, unsigned int maxlist) {
  unsigned int i, ilist, j;
  float x;

  for(i=0; i<maxlist*3; i++) {
    ilist = i%maxlist;
    j = (unsigned int)floor(randomx()*maxlist);
	  x = list[ilist];
    list[ilist] = list[j];
    list[j] = x;
  }
}


// skip to end of line, including end of line character
void ignoreline(FILE *ifp) {
  fscanf(ifp, "%*[^\n]");
  fscanf(ifp, "%*1[\n]");
}

// convert double to float while preventing underflow
float d2f(double val) {
  if(val<0.0) return -val<FLT_MIN?0.0f:(float)val;
  return val<FLT_MIN?0.0f:(float)val;
}

// return sqrt in float format while preventing underflow
float sqrtdf(double val) {
  return d2f(sqrt(val));
}

// convert a string to float while preventing underflow
float a2f(char * s) {
  return d2f(atof(s));
}



// next 3 functions: table-based sqrt and 1/sqrt evaluation
// tables are computed for the sqrt and 1/sqrt of mantissa and exponent separately
// (one table, the shortest one related to the exponent, could be eliminated by simply 
//  bit-shifting the exponent by one bit... to be explored)

// fill tables
void mySqrtTableFill() {
  unsigned long i, j, n = 1<<16; // 1<<16: mantissa tables contain 65536 values
  float val;

  // mantissa from 0 to 2 in 65536 steps, giving off precision!
  for(i=0;i<n;i++) {
    // generate a float values 
    j = (i<<7) + ((1<<23)*127); 
    val = *(float *)&j;
    // store its sqrt and 1/sqrt in tables
    mySqrtTable[i]    = sqrtdf(val);  
    myInvSqrtTable[i] = 1.0f/mySqrtTable[i];
  }

  // exponent from 2^-128 to 2^128
  for(i=0;i<(1<<8);i++) {
    // generate a float values 
    j = i<<23;
    val = *(float *)&j;
    // store its sqrt and 1/sqrt in tables
    mySqrtTableExp[i]    = sqrtdf(val);
    myInvSqrtTableExp[i] = 1.0f/mySqrtTableExp[i];
  }
}


// sqrt of val is the product of the sqrt of its mantissa and sqrt of its exponent
// WARNING: approximate solution, precise to 0.0015%, use when precision is not critical
float mySqrt(float val) {
  if( *(unsigned long *)&val==0 ) return 0.0f;
  return mySqrtTable[((*(unsigned long *)&val)>>7)&0xFFFF]*mySqrtTableExp[((*(unsigned long *)&val)>>23)&0xFF];
}

// 1/sqrt of val is the product of the 1/sqrt of its mantissa and 1/sqrt of its exponent
// WARNING: approximate solution, precise to 0.0015%, use when precision is not critical
float myInvSqrt(float val) {
  if((*(unsigned long *)&val)==0) return 1.0f/val; // prevent division by 0
  return myInvSqrtTable[((*(unsigned long *)&val)>>7)&0xFFFF]*myInvSqrtTableExp[((*(unsigned long *)&val)>>23)&0xFF];
}

